package icommand.robotics;

import icommand.platform.nxt.*;

//!! Need to take into account SyncMotors.inverted;

// !! For rotation, it should straighten out when stopped to try to keep
// the angle somewhat accurate.

// !! Should all methods call stop() first in case it was roaming?
// OR methods account for RCX currently in moving mode?

// !! All methods that change x, y must be synchronized.

/**
 * The ServoNavigator class contains methods for performing basic navigational
 * movements. This class uses two standard NXT servo motors to monitor the wheels of the
 * differential drive. <BR>
 * PRELIMINARY: Doesn't work very good because of SyncMotors poor performance.
 * Note: This class will only work for robots using two motors to steer differentially
 * that can rotate within its footprint (i.e. turn on the spot).
 * 
 * @author <a href="mailto:bbagnall@mts.net">Brian Bagnall</a>
 * @version 0.1  22-October-2006
 */
public class ServoNavigator implements Navigator {

   // orientation and co-ordinate data
   private double angle;
   private float x;
   private float y;

   // Synchronized Motors:
   private SyncMotors vehicle;
   
   // Movement values:
   private double COUNTS_PER_CM;
   private double COUNTS_PER_DEGREE;
   
   // Internal states
   private boolean moving;
   
   /**
    * For use only if the wheel axle is inserted directly into the motor.
    * (i.e. use a different constructor if your drive train uses gears)
    * @param wheelDiameter
    * @param driveLength
    * @param vehicle
    */
   public ServoNavigator(double wheelDiameter, double driveLength, SyncMotors vehicle) {
	   this(wheelDiameter, driveLength, 360, vehicle);
   }
   
   /**
   * Allocates a ServoNavigator object and initializes it with a SyncMotors object.
   * The x and y values will each equal 0 (cm's) on initialization, and the starting
   * angle is 0 degrees, so if the first move is forward() the robot will run along
   * the x axis. <BR>
   * @param wheelDiameter The diameter of the wheel, usually printed right on the
   * wheel, in millimeters (e.g. 56 mm) 
   * @param driveLength The distance from the center of the left tire to the center
   * of the right tire, in centimeters.
   * @param ratio The ratio of sensor rotations to wheel rotations.<BR>
   *  e.g. 720 tacho counts for every one turn of the wheel = 720f<BR>
   * If your wheel axle is directly in the motor, use 360. 
   * @param vehicle A SyncMotors object containing two synchronized motors.
   */
   public ServoNavigator(double wheelDiameter, double driveLength, double ratio, SyncMotors vehicle) {
      this.vehicle = vehicle;
      
      // Set coordinates and starting angle:
      angle = 0.0;
      x = 0.0f;
      y = 0.0f;
      
      moving = false;
      
      // Calculate the counts per centimeter
      double wheelCircumference = (wheelDiameter/10) * Math.PI;
      COUNTS_PER_CM = ratio/wheelCircumference;
      //System.out.println("Counts/cm = " + COUNTS_PER_CM);
      
      // Calculate counts per degree
//      double fullRotation = (driveLength * Math.PI);
//      COUNTS_PER_DEGREE = ((fullRotation/wheelCircumference)*ratio)/360;
      COUNTS_PER_DEGREE = COUNTS_PER_CM * ((2 * Math.PI * driveLength)/360);
      //System.out.println("Counts/degree = " + COUNTS_PER_DEGREE);
   }
   
   /**
    * 
    * @param wheelDiameter The diameter of the wheel, usually printed right on the
    * wheel, in millimeters (e.g. 56 mm) 
    * @param driveLength The distance from the center of the left tire to the center
    * of the right tire, in centimeters.
    * @param ratio The ratio of sensor rotations to wheel rotations.<BR>
    *  e.g. 3 complete rotations of the sensor for every one turn of the wheel = 3f<BR>
    * 1 rotation of the sensor for every 2 turns of the wheel = 0.5f
    * @param left Left Motor e.g. Motor.B
    * @param right Right Motor e.g. Motor.B
    */
   public ServoNavigator(float wheelDiameter, float driveLength, float ratio, Motor left, Motor right) {
      this(wheelDiameter, driveLength, ratio, new SyncMotors(left, right));
   }
   
   /**
   * Returns the current x coordinate of the NXT.
   * If NXT is moving it will try to calculate the current position.
   * @return float Present x coordinate.
   */
   public float getX() {
      // Get tacho count, add to global X but don't change global X.
	  if(moving) {
	      double centimeters = vehicle.left.getBlockTacho() / COUNTS_PER_CM;
	   	  return x + (float)(Math.cos(Math.toRadians(getAngle())) * centimeters);
	  } else
		  return x;
   }
   
   /**
   * Returns the current y coordinate of the NXT.
   * If NXT is moving it will try to calculate the current position.
   * @return float Present y coordinate.
   */
   public float getY() {
		  if(moving) {
		      double centimeters = vehicle.left.getBlockTacho() / COUNTS_PER_CM;
		   	  return y + (float)(Math.sin(Math.toRadians(getAngle())) * centimeters);
		  } else
			  return y;
   }

   /**
   * Returns the current angle the NXT robot is facing.
   * Note: At present it will only give an updated reading when the NXT is not in the midst of rotating.
   * @return float Angle in degrees.
   */
   public double getAngle() {
      return angle;
   }

   /**
   * Rotates the NXT robot a specific number of degrees in a direction (+ or -).
   * This method will return once the rotation is complete.
   * PRELIMINARY: The code is accurate but the SyncMotors.left(count) and
   * SyncMotors.right(count) are not very good right now. 
   *
   * @param degrees Angle to rotate in degrees. A positive value rotates left, a negative value right.
   */
   public void rotate(double degrees) {
      
      // keep track of angle
      this.angle = this.angle + degrees;
      this.angle = (int)this.angle % 360; // Must be < 360 degrees
      
      // Is it possible to do the following with modulo (%) ???
      while(this.angle < 0)
         this.angle += 360;  // Must be > 0
      
      // Calculate the number of intervals of rotation sensor to count
      int count = (int)(COUNTS_PER_DEGREE * degrees);
            
      if (degrees > 0) {
    	  vehicle.left(count, false);
      }
      else if (degrees < 0) {
    	  vehicle.right(Math.abs(count), false);
      }
   }

   /**
   * Rotates the NXT robot to point in a certain direction. It will take the shortest
   * path necessary to point to the desired angle. Method returns once rotation is complete.
   *
   * @param angle The angle to rotate to, in degrees.
   */
   public void gotoAngle(double degrees) {
      // in future, use modulo instead of while loop???
      double difference = degrees - this.angle;
      while(difference > 180)
         difference = difference - 360; // shortest path to goal angle
      while(difference < -180)
         difference = difference + 360; // shortest path to goal angle
      rotate(difference);
   }

   /**
   * Rotates the NXT robot towards the target point and moves the required distance.
   *
   * @param x The x coordinate to move to.
   * @param y The y coordinate to move to.
   */
   public void gotoPoint(double x, double y) {

      // Determine relative points
      double x1 = x - this.x;
      double y1 = y - this.y;

      // Calculate angle to go to:
      float angle = (float)Math.atan2(y1,x1);

      // Calculate distance to travel:
      double distance;
      if(y1 != 0)
         distance = y1/Math.sin(angle);
      else
         distance = x1/Math.cos(angle);

      // Convert angle from rads to degrees:
      angle = (float)Math.toDegrees(angle);
      
      // Now convert theory into action:
      gotoAngle(angle);
      travel(Math.round(distance));
   }

   /**
   * Moves the NXT robot a specific distance. A positive value moves it forwards and
   * a negative value moves it backwards. Method returns when movement is done.
   * PRELIMINARY (Currently overshoots by about half an axle turn)
   *
   * @param dist The positive or negative distance to move the robot (in centimeters).
   */
   public void travel(long dist) {
	   // Do I want this to return immediately? Pretty easy to make it
	   // return and make it able to keep track of count even if interrupted.
      int counts = (int)(dist * COUNTS_PER_CM);
      //System.out.println("COUNTS_PER_CM = " + COUNTS_PER_CM);
      //System.out.println("Counts = " + counts);
      this.moving = true;
      if(dist > 0) {
         vehicle.forward(Math.abs(counts));
      } else
      if(dist < 0) {
         vehicle.backward(Math.abs(counts));
      }
      this.moving = false;
      recalculate();
   }

   /**
   * Moves the NXT robot forward until stop() is called.
   *
   * @see Navigator#stop().
   */
   public void forward() {
	   vehicle.forward();
	   this.moving = true;
   }

   
   /**
   * Moves the NXT robot backward until stop() is called.
   *
   * @see Navigator#stop().
   */
   public void backward() {
	   vehicle.backward();
	   this.moving = true;
   }

   /**
   * Halts the NXT robot and calculates new x, y coordinates.
   *
   * @see Navigator#forward().
   */
   public void stop() {
      if(moving) {
         vehicle.stop();
    	  
         recalculate();
      }
   }
   
   /**
    * Recalculate x-y coordinates based on servo tacho count
    *
    */
   private void recalculate() {
	   
       int rotAvg = (vehicle.left.getBlockTacho() + vehicle.right.getBlockTacho()) / 2;
       double centimeters = rotAvg / COUNTS_PER_CM;
    
       // update x, y coordinates
       x = x + (float)(Math.cos(Math.toRadians(angle)) * centimeters);
       y = y + (float)(Math.sin(Math.toRadians(angle)) * centimeters);
   }
}
